#!/usr/bin/python
# -*- coding: utf-8 -*-
'''

  product:  LinOTP2
  module:   Admin Client
  tool:     linotpadm.py
  edition:  Community Edition
  
  Copyright (C) 2010 - 2013 LSE Leading Security Experts GmbH
  License:  AGPLv3
  contact:  http://www.linotp.org
            http://www.lsexperts.de
            linotp@lsexperts.de

  Description:  This is the admin client to manage the tokens
                on the LinOTP 2 server
                 
  Dependencies: clientutils, 
                etokenng (for EE)

'''

VERSION='2.4.3'
import os
from getopt import getopt, GetoptError
import getpass
from linotputils.clientutils import *
from linotputils.yubikey import enrollYubikey
from linotputils.yubikey import YubikeyPlug
import pprint

try:
    from linotputilsee.etokenng import *
    EDITION="EE"
except ImportError:
    print 
    print "Thank you for running the LinOTP2 linotpadmin client Community Edition"
    print
    EDITION="CE"

REALM = "LinOTP2 admin area (admin/admin)"
COMMANDS = [ 'listtoken', 'inittoken', 'assigntoken', 
            'importtoken', 'disabletoken', 'enabletoken', 
            'resynctoken', 'removetoken', 'set',
            'unassigntoken', 'listuser', 'getconfig', 'setconfig',
            'getrealms', 'setrealm', 'deleterealm', 'setdefaultrealm',
            'getresolvers', 'deleteresolver', 'setresolver',  'loadtokens', 
            'yubikey_mass_enroll', 'etokenng_mass_enroll',
            'securitymodule' ]

def usage():
    print "usage: %s --url=<url> --admin=<adminusername> --cert=<cert> --key=<rsakey> --command=<command> --version" % sys.argv[0]
    print "  --url/-U     : The base url of the LinOTP server. Something like"
    print "                 http://localhost:5000 or https://linotp:443"
    print "  --admin/-a   : If the admin interface of the LinOTP service requires authentication"
    print "                 you need to pass the username and will be asked for the password."
    print "  --authtype   : The default authtype is 'Digest'. You may change this to 'Basic'"
    print "  --cert/-c    : If the admin interface of the LinOTP service requires authentication via client certificate"
    print "                 you may pass a P12 file here."
    print "  --key/-k     : The private key for authenticating"
    print "  --admin/-a   : The name of the administrative account, when authenticating with username/password"
    print "  --version/-v : Print version"
    print "  --help/-h    : Print this help screen"
    print
    print "  --command/-C :"
    print "    listtoken:    [--user | --serial ]"
    print "    listuser:     "
    if EDITION == "EE":
        print "    inittoken:    --user= --serial= --desecription= --pin= --otpkey= --etng --pytoken --type=<type>"
    else:
        print "    inittoken:    --user= --serial= --description= --pin= --otpkeyc="
        
    print "    yubikey_mass_enroll"
    print "        --yubiprefix=<string>"
    print "        --yubiprefixrandom=<length>"
    print "        --yubiprefixserial          : use the serial of the yubikey as prefix"
    print "        --yubimode=<OATH or YUBICO>"
    print "        --yubislot=<1 or 2>"
    print "    etokenng_mass_enroll [--label=TokenName]"
    print "    assigntoken:    --user --serial"
    print "    unassigntoken:  --serial"
    print "    importtoken:    --file or -f "
    print "    disabletoken:   --serial | --user"
    print "    enabletoken:    --serial | --user"
    print "    removeetoken:   --serial | --user"
    print "    resynctoken:    --serial | --user   --otp1  --otp2"
    print "    set:       [--user | --serial ] --pin --maxfailcount --syncwindow --otplen"
    print 
    print "    Security Modules"
    print "    ----------------"
    print "    securitymodule:         get the status of the security module or set the password for the module"
    print "                            To set the password, you must specify the following parameter:"
    print "                                   --module=<modulename>"
    print 
    print "    Server configuration"
    print "    --------------------"
    print "    getconfig:    returns the configuration of the LinOTP server"
    print "    setconfig:    sets a certain config value like this"
    print "                  --config='DefaultSyncWindow=500'"
    print "    Realm and resolver configuration"
    print "    --------------------------------"
    print "    getrealms:               lists all realms"
    print "    setrealm:        --realm=<realmname> --resolver=<resolverlist>"
    print "    deleterealm:     --realm=<realmname>"
    print "    setdefaultrealm: --realm=<realmname>"
    print "    getresolvers:            returns a list of all available resolvers"
    print "    deleteresolver:  --resolver=<resolvername>"
    print "    setresolver:     --resolver=<> --rtype=[LDAP,SQL,FILE] ...."
    print
    print "  Paramters for commmands:"
    print "    --user=<user> or -u <user>"
    print "    --pin=<pin> or -p <pin>"
    print "    --serial=<S/N> or -s <serial>"
    print "    --description=<description for Token> or -d <description>"
    print "    --otpkey=<HMAC key> or -H  "
    print "                      This is usually not necessary, as keys are read from the file"
    print "    --file=<file with Token data> or -f <file>"
    print "    --otpval=<otp value for resyncing>"
    print "    --window=<resyncing windows size> or -w"
    print "    --etng or -e     "
    print "                      This parameter takes no value. If it is passed to the inittoken command, "
    print "                      an Aladdin eTokenNG OTP is initialized. --hmac and --serial are ignored."    
    print "    --pytoken         This parameter takes no value. This will create a soft Token as python script"
    print "                      example: linotpadm.py -U https://localhost -a admin -C inittoken --type=HMAC --pytoken --user=jdoe"
    print "    --type=<type> or -t  "
    print "                      This specifies the Type of the OTP Token to be initialized."
    print "                      HMAC is default. You may specify the type 'motp'"
    print "                      When using 'motp' you also need to specify --otppin="
    print "    --realm=<realmname> "
    print "                      gives the realm to operate on."
    print "    --resolver=<resolvername> or <resolverlist>"
    print "                      comma seperated list of resolvers or"
    print "                      single resolver"
    print 
    print "  Parameters for the command setresolver:"
    print "    --rtype           Specify the type of the resolver: LDAP, SQL or FILE"
    print "                      Depending on the resolver type, there are different parameters to set."
    print "    --rf_file         Resolver FILE: the filename containing the users on the LinOTP server"
    print "    --rl_uri          Resolver LDAP: the LDAP uri of the ldap server"
    print "    --rl_basedn       Resolver LDAP: the BaseDN for searching for users"
    print "    --rl_binddn       Resolver LDAP: the BindDN to authenticate to the LDAP server"
    print "    --rl_bindpw       Resolver LDAP: the Bind Password to authenticate to the LDAP server"
    print "    --rl_timeout      Resolver LDAP: the timeout when connecting to the LDAP server"
    print "    --rl_loginattr    Resolver LDAP: the attribute containing the loginname like uid"
    print "    --rl_searchfilter Resolver LDAP: the searchfilter. Something like (uid=*)(objectClass=inetOrgPerson)"
    print "    --rl_userfilter   Resolver LDAP: the userfilter for reverse resoling the DN for a given loginname,"
    print "                                     something like (&(uid=%s)(ObjectClass=inetOrgPerson))"
    print "    --rl_attrmap      Resolver LDAP: the attribute mapping. Something like:"
    print '                                     { "username": "uid", "phone" : "telephoneNumber", "groups" : "o", "mobile" : "mobile", "email" : "mail", "surname" : "sn", "givenname" : "givenName" }'
    print "    TODO: All the SQL parameters... "
    



def showresult( rv ):
    pp = pprint.PrettyPrinter(indent=4)
    print pp.pformat( rv['result'] )
    
    
##### main

def main():
    url = None
    command = None
    file = None
    admin = None
    adminpw = None
    param = {}
    etng = False
    pytoken = False
    certificate=None
    key=None
    protocol = "http"
    authtype = "Digest"
    yubi_prefix = None
    yubi_prefix_random = 0
    yubi_mode = 2 # OATH
    yubi_slot = 1
    yubi_prefix_serial = False

    try:
        opts, args = getopt(sys.argv[1:], "hvU:a:C:k:c:f:u:s:p:d:ew:t:w:m:H:", 
                ["help", "version", "url=", "admin=", "cert=", "key=", "command=", "file=", 
                "user=", "serial=", "pin=", "otpkey=", "description=", "etng", "maxfailcount=", 
                "syncwindow=", "otplen=", "otp1=", "otp2=", "pytoken", "type=", "otppin=", 
                "config=",
                'realm=', 'resolver=', 'rtype=',
                'module=',
                'label=', 'authtype=',
                'yubiprefix=', 'yubiprefixrandom=', 'yubimode=', 'yubislot=',
                'yubiprefixserial'] + file_opts + ldap_opts)
        
    except GetoptError:
        print "There is an error in your parameter syntax:"
        usage()
        sys.exit(1)

##### Creating Parameter list

    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage()
            sys.exit(0)
        elif opt in ('-U', '--url'):
            if arg.startswith('https://'):
                protocol="https" 
                url = arg[8:]
            elif arg.startswith('http://'):
                protocol="http"
                url =arg[7:]
            else:
                print "Malformed url format. You need to start with http or https ["+arg+"]"
                sys.exit(1)
        elif opt in ('-a', '--admin'):
            admin = arg
            adminpw = getpass.getpass( prompt="Please enter password for '%s':" % admin) 
        elif opt in ('-c', '--cert'):
            certificate = arg
        elif opt in ('-k', '--key'):
            key = arg
        elif opt in ('-C', '--command'):
            command = arg
        elif opt in ('-f', '--file'):
            file = arg
        # user, serial (needed for assigntoken)
        elif opt in ('-u', '--user'):
            param['user'] = arg
        elif opt in ('-s', '--serial'):
            param['serial'] = arg
        # pin, hmac, description (needed for inittoken)
        elif opt in ('-p', '--pin'):
            param['pin'] = arg
        elif opt in ('-H', '--otpkey'):
            param['otpkey'] = arg
        elif opt in ('-m','--maxfailcount'):
            param['maxfailcount'] = arg
        elif opt in ('-w','--window'):
            param['syncwindow']=arg
        elif opt in ('--otplen'):
            param['otplen'] = arg
        elif opt in ('--otp1'):
            param['otp1'] = arg
        elif opt in ('--otp2'):
            param['otp2'] = arg
        elif opt in ('-d', '--description'):
            param['description'] = arg
        elif opt in ('--label'):
            param['label'] = arg
        elif opt in ('--etng'):
            if EDITION=="EE":
                etng = True
        elif opt in ('--pytoken'):
            pytoken = True
        elif opt in ('-t','--type'):
            param['type']= arg
        elif opt in ('--otppin'):
            param['otppin'] = arg
        elif opt in ('-v','--version'):
            print "linotpadm.py "+VERSION
            sys.exit(0)
        elif opt in ('--config'):
            key, value=arg.rsplit('=',2)
            param[key]=value
        elif opt in ('--realm'):
            param['realm']= arg
        elif opt in ('--resolver'):
            param['resolver']=arg
        elif opt in ('--rtype'):
            param['rtype'] = arg
            
        elif opt in ('--rf_file'):
            param['rf_file'] = arg
        elif opt in ('--authtype'):
            authtype = arg
        elif opt in ('--module='):
            param['module'] = arg
        elif opt in ('--yubiprefix='):
            yubi_prefix = arg
        elif opt in ('--yubiprefixrandom='):
            yubi_prefix_random = int(arg)
        elif opt in ('--yubimode='):
            if arg.lower() == "yubico":
                yubi_mode=1
        elif opt in ('--yubislot='):
            yubi_slot = int(arg)
        elif opt in ('--yubiprefixserial'):
            yubi_prefix_serial = True
    
        for l_opt in ldap_opts:
            o = l_opt.split('=')
            if opt in ( "--%s" % o[0] ):
                param[o[0]] = arg        

    if not (url and command):
        print "Missing arguments. See --help"
        usage()
        sys.exit(0)

    if command not in COMMANDS:
        print "Unknown command. See --help"
        usage()
        sys.exit(0)

    # Create the linotpclient instance
    lotpc = linotpclient( protocol, url, admin=admin, adminpw=adminpw, cert=certificate, key=key, authtype=authtype)

##### The commands
    
    if ( command == "listtoken" ):
        r1 = lotpc.listtoken(  param  )
        result=r1['result']
           
        tabsize    =[4,16,12,20,20,4,4,4,4]
        tabstr     =["%4s","%16s","%12s","%20s","%20s","%4s","%4s","%4s","%4s","%4s"]
        tabdelim   ='|'
        tabvisible =[0,1,2,3,4,5,6,7,8]
        tabhead    =['Id','Desc','S/N','User','Resolver','MaxFail','Active','FailCount','Window']
        tabentry    =['LinOtp.TokenId',
                     'LinOtp.TokenDesc',
                     'LinOtp.TokenSerialnumber',
                     'User.username',
                     'LinOtp.IdResClass',                     
                     'LinOtp.MaxFail',
                     'LinOtp.Isactive',
                     'LinOtp.FailCount',
                     'LinOtp.SyncWindow']
        dumpresult( result['status'], result['value']['data'], { 'tabsize' : tabsize, 'tabstr' : tabstr, 
                    'tabdelim' : tabdelim, 'tabvisible': tabvisible,
                    'tabhead' : tabhead,'tabentry' : tabentry } )    
    elif ( command == "listuser" ):
        r1 = lotpc.userlist(  {'username':'*'} )
        result = r1['result']
        tabentry    =['username',
                     'surname',
                     'userid',
                     'phone',
                     'mobile',
                     'email',
                     ]
        tabsize    =[20,20,20,20,20,20]
        tabstr     =["%20s","%20s","%20s","%20s","%20s","%20s"]
        tabdelim   ='|'
        tabvisible =[0,1,2,3,4,5]
        tabhead    =['login','surname','Id','phone','mobile','email']
        dumpresult( result['status'], result['value'], { 'tabsize' : tabsize, 'tabstr' : tabstr, 
                    'tabdelim' : tabdelim, 'tabvisible': tabvisible,
                    'tabhead' : tabhead,'tabentry' : tabentry } )    
    elif ( command == "getconfig" ):
        r1 = lotpc.readserverconfig( {} )
        showresult(r1)
    elif ( command == "setconfig" ):
        print param
        r1 = lotpc.writeserverconfig( param )
        showresult(r1)
    elif ( command == "assigntoken" ):
        if 'user' not in param or 'serial' not in param :
            print "To assign a token, we need a username and a tokenserial:"
            print "   --command=assigntoken --user=<username> --serial=<tokenserial>"
            sys.exit(1)
        r1 = lotpc.assigntoken(  param)
        showresult(r1)
    elif ( command == "unassigntoken" ):
        if 'serial' not in param :
            print "To unassign a token, we need a tokenserial:"
            print "   --command=unassigntoken --serial=<tokenserial>"
            sys.exit(1)
        r1 = lotpc.unassigntoken( param)
        showresult(r1)
    elif ( command == "yubikey_mass_enroll"):
        yp = YubikeyPlug()
        while 0==0:
            print "\nPlease insert the next yubikey... " 
            #input = raw_input("Please insert the next yubikey and press enter (x=Exit): ")
            #if "x" == input.lower():
            #    break
            ret = yp.wait_for_new_yubikey()
            otpkey, serial = enrollYubikey( debug= False,
                                            prefix_serial = yubi_prefix_serial,
                                            fixed_string = yubi_prefix,
                                            len_fixed_string = yubi_prefix_random,
                                            slot = yubi_slot,
                                            mode = yubi_mode )
            description = param.get('description', "mass enrolled")
            if yubi_mode == 2:
                # According to http://www.openauthentication.org/oath-id/prefixes/
                # The OMP of Yubico is UB
                # As TokenType we use OM (oath mode)
                r1 = lotpc.inittoken( {'serial':"UBOM%s"%serial, 
                                       'otpkey':otpkey, 
                                       'description':description} )
                showresult(r1)
            elif yubi_mode == 1:
                yubi_otplen = 32
                if yubi_prefix_serial:
                    yubi_otplen = 32 + len(serial)*2
                elif yubi_prefix:
                    yubi_otplen = 32 + (len(yubi_prefix)*2)
                elif yubi_prefix_random:
                    yubi_otplen = 32 + (yubi_prefix_random*2)
                # According to http://www.openauthentication.org/oath-id/prefixes/
                # The OMP of Yubico is UB
                # As TokenType we use AM (AES mode)
                r1 = lotpc.inittoken( {'type' : 'yubikey', 
                                       'serial':"UBAM%s"%serial, 
                                       'otpkey':otpkey, 
                                       'otplen':yubi_otplen,
                                       'description':description} )
                showresult(r1)
            
    elif ( command == "etokenng_mass_enroll" ):
        print "Mass-Enrolling eToken NG OTP. Beware the tokencontents of all tokens will be deleted."
        print "Random User PINs and SO-PINs will be set. The SO-PIN will be stored in the Token-Datenbase."
        print
        while 0==0:
            input = raw_input("Please insert the next eToken NG and press enter (x=Exit): ")
            if "x" == input.lower():
                break
            tokenlabel = param.get('label', "LinOTPToken")
            description = param.get('description', 'eTokenNG mass enrolled')
            tdata  = initetng({ 'label': tokenlabel, 'debug' : False,
                               'description' : description })
            if not tdata['userpin']  or not tdata['hmac'] or not tdata['serial']:
                print "No token was added to LinOTP:", tdata['error']
                sys.exit(1)
            param['serial']=tdata['serial']
            param['otpkey']=tdata['hmac']
            param['userpin']=tdata['userpin']
            param['sopin']=tdata['sopin']
            r1 = lotpc.inittoken( param)
            showresult(r1)
        
    elif ( command == "inittoken" ):
        if etng:          
            tokenlabel = param.get('user', param.get('label', "LinOTPeToken"))
            tdata  = initetng({ 'label': tokenlabel, 'debug' : True })
            if not tdata['userpin']  or not tdata['hmac'] or not tdata['serial']:
                print "No token was added to LinOTP:", tdata['error']
                sys.exit(1)
            param['serial']=tdata['serial']
            param['otpkey']=tdata['hmac']
            param['userpin']=tdata['userpin']
            param['sopin']=tdata['sopin']
            print "FIXME: what shall we do with the eToken password and SO PIN:", tdata['userpin'], tdata['sopin']
        elif pytoken:
            if 'user' not in param:
                print "To initialize a pyToken, please provide a username"
                sys.exit(1)
            pyTemplate="FAIL"
            pyTemplateList= ( 'pytoken.template.py',
                    '/usr/share/pyshared/linotputils/pytoken.template.py',
                    '/usr/local/lib/python2.6/dist-packages/linotputils/pytoken.template.py',
                    '/usr/lib/python2.6/dist-packages/linotputils/pytoken.template.py',
                )
            for pT in pyTemplateList:
                if os.path.isfile(pT):
                    pyTemplate=pT
                    break
            if pyTemplate == "FAIL":
                print "Could not find any pytoken template!"
                sys.exit(1)
            else:
                pyTok = pyToken(keylen=256, template=pyTemplate)
                pyTokenfile=pyTok.createToken( param['user'] )
                param['otpkey']=pyTok.getHMAC()
                param['serial']=pyTok.getSerial()
                print pyTokenfile
        else:
            if 'user' not in param or 'serial' not in param or 'otpkey' not in param:
                print "To initialize a token, we need at least a username, a tokenserial and an OTPkey/HMAC:"
                print "   --command=inittoken --user=<username> --serial=<tokenserial> --hmac=<HMAC>"
                sys.exit(1)
        r1 = lotpc.inittoken( param)
        showresult(r1)
    elif ( command == "importtoken" ):
        if not file:
            print "To import tokens, we need a filename!"
            sys.exit(1)
        lotpc.importtoken ( { 'file': file } )
    elif ( command == "disabletoken" ):
        r1 = lotpc.disabletoken (  param)
        showresult (r1)
    elif ( command == "removetoken" ):
        r1 = lotpc.removetoken (  param )
        showresult (r1)
    elif ( command == "enabletoken" ):
        r1 = lotpc.enabletoken ( param)
        showresult (r1)
    elif ( command == "resynctoken" ):
        r1 = lotpc.resynctoken ( param)
        showresult (r1)
    elif ( command == "set" ):
        if 'user' not in param and 'serial' not in param:
            print "Please provide either or user or a serial to set the pin."
            sys.exit(1)
        r1 = lotpc.set(param)
        showresult(r1)
    elif ( command == "getrealms" ):
        r1 = lotpc.getrealms( {})
        showresult (r1)
    elif ( command == "setrealm" ):
        if "realm" not in param or "resolver" not in param:
            print "You need to provide a realm and resolvers."
            sys.exit(1)
        param['resolvers']=param['resolver']
        r1 = lotpc.setrealm( param)
        showresult (r1)        
    elif ( command == "deleterealm" ):
        if "realm" not in param:
            print "You need to provide a realm."
            sys.exit(1)   
        r1 = lotpc.deleterealm(param)
        showresult (r1)
    elif ( command == "securitymodule" ):
        r1 = {}
        if "module" in param:
            password = getpass.getpass( prompt="Please enter password for security module '%s':" % param["module"])
            print "Setting the password of your security module %s" % param['module']  
            r1 = lotpc.securitymodule(param = { "hsm_id" : param["module"],
                                                "password" : str(password) })
        else:
            print "This is the configuration of your active Security module:"
            print
            r1 = lotpc.securitymodule(param={})
        showresult(r1) 
    elif ( command == "setdefaultrealm" ):
        if "realm" not in param:
            print "You need to provide a realm."
            sys.exit(1)
        r1 = lotpc.setdefaultrealm(param)
        showresult (r1)        
    elif ( command == "getresolvers" ):
        r1 = lotpc.getresolvers( {} )
        showresult (r1)        
    elif ( command == "deleteresolver" ):
        if "resolver" not in param:
            print param
            print "You need to provide a resolver."
            sys.exit(1)        
        r1 = lotpc.deleteresolver( param )
    elif ( command == "setresolver" ):
        if not "rtype" in param:
            print "your need to specify a type of the resolver."
            sys.exit(1)
        if param['rtype'] == 'LDAP':
            for opt in ldap_opts:
                o=opt.split('=')
                if o[0] not in param:
                    print "you need to specify --%s for ldap resolvers" % o[0]
                    sys.exit(1)
        elif param['rtype'] == 'SQL':
            print "TODO: SQL parameters not implemented yet"
            sys.exit(1)
        elif param['rtype'] == 'FILE':
            if 'rf_file' not in param:
                print "you need to specify --rf_file!"
                sys.exit(1)
        r1 = lotpc.setresolver( param )
        showresult(r1)
    elif ( command == "loadtokens" ):
        filename = 'plain-value-pskc.xml'
        f = open( filename, 'r')
        xml_data = f.read()
        ret = lotpc.connect( "/admin/loadtokens", {}, 
                    { 'type' : 'pskc', 
                        'file' : xml_data, 
                        'pskc_type' : 'plain', 
                        'pskc_password' : None, 
                        'pskc_preshared': None })
        showresult(ret)

    else:
        print "Nothing to do and nothing done."
 

if __name__ == '__main__':
    main()

